import json
import cherrypy
from framepy import core
from framepy import modules
from framepy import _method_inspection

DEFAULT_ENCODING = 'utf-8'
annotated_controllers = {}


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        self._map_controllers_from_arguments(arguments)

    def after_setup(self, properties, arguments, context, beans_initializer):
        controllers_mappings = [core.Mapping(controller(), key) for key, controller in annotated_controllers.items()]

        for m in controllers_mappings:
            beans_initializer.initialize_bean('__controller_'.format(m.bean.__class__.__name__), m.bean, context)
            cherrypy.tree.mount(m.bean, m.path)

    def _map_controllers_from_arguments(self, arguments):
        controllers = arguments.get('controllers') or []
        for controller_class, path in controllers:
            annotated_controllers[path] = controller_class


def controller(path):
    def wrapped(potential_controller_class):
        annotated_controllers[path] = potential_controller_class
        return potential_controller_class
    return wrapped


class PayloadEntity(object):
    def __init__(self, entries):
        self.__dict__.update(entries)


class ResponseEntity(object):
    class JsonEncoder(json.JSONEncoder):
        def default(self, o):
            return o.__dict__
    _json_encoder = JsonEncoder()

    def __init__(self, status='ok', data=None, error=None):
        self.status = status
        if data is not None:
            self.data = data
        if error is not None:
            self.error = error

    def tojson(self):
        return ResponseEntity._json_encoder.encode(self).encode(DEFAULT_ENCODING)


class PayloadConstraint(object):
    def __init__(self, name, required=True, type='string', min=0, max=0xffffffff, default=None, nested=None):
        self.name = name
        self.required = required
        self.type = type
        self.min = min
        self.max = max
        self.default = default
        self.nested = nested


class PayloadBinder(object):
    def __init__(self, payload, payload_template):
        self.types_resolvers = {'int': self._resolve_int, 'float': self._resolve_float, 'string': self._resolve_string}
        self.errors = []
        self._payload_template = payload_template
        self._bind_payload(payload)

    def has_errors(self):
        return len(self.errors) > 0

    def _bind_payload(self, payload):
        constraints = [getattr(self._payload_template, constraint)
                       for constraint in self._payload_template.__dict__
                       if isinstance(getattr(self._payload_template, constraint), PayloadConstraint)]
        fields = {}
        for constraint in constraints:
            value = payload.get(constraint.name, None)
            value_to_bind = self._process_nested_payload(value, constraint) if constraint.nested is not None \
                else self._process_fields(value, constraint)

            if value_to_bind is not None:
                fields[constraint.name] = value_to_bind

        self.entity = PayloadEntity(fields)

    def _process_nested_payload(self, value, constraint):
        if value is None and constraint.required:
            self.errors.append({'field': constraint.name, 'error': 'MISSING_FIELD'})
            return None
        elif value is not None:
            another_binder = PayloadBinder(value, constraint.nested)
            if another_binder.has_errors():
                self.errors.extend(another_binder.errors)
            return another_binder.entity

    def _process_fields(self, value, constraint):
        if value is None and constraint.required:
            self.errors.append({'field': constraint.name, 'error': 'MISSING_FIELD'})
            return None
        elif value is None and not constraint.required:
            value = constraint.default

        if value is None:
            return None

        try:
            self._resolve_field(constraint, value)
        except ValueError:
            self.errors.append({'field': constraint.name, 'error': 'BAD_TYPE'})
        return value

    def _resolve_field(self, constraint, value):
        type_resolver = self.types_resolvers.get(constraint.type)
        if type_resolver is not None:
            type_resolver(value, constraint)
        else:
            raise UnknownFieldType('Unknown field type {0}'.format(constraint.type))

    def _resolve_int(self, value, constraint):
        self._check_type_and_ranges(value, constraint,
                                    type_conversion_procedure=lambda val: int(val),
                                    size_check_procedure=lambda val: val,
                                    max_error_label='MAX',
                                    min_error_label='MIN')

    def _resolve_float(self, value, constraint):
        self._check_type_and_ranges(value, constraint,
                                    type_conversion_procedure=lambda val: float(val),
                                    size_check_procedure=lambda val: val,
                                    max_error_label='MAX',
                                    min_error_label='MIN')

    def _resolve_string(self, value, constraint):
        self._check_type_and_ranges(value, constraint,
                                    type_conversion_procedure=lambda val: str(val),
                                    size_check_procedure=lambda val: len(val),
                                    max_error_label='MAX_LENGTH',
                                    min_error_label='MIN_LENGTH')

    def _check_type_and_ranges(self, value, constraint,
                               type_conversion_procedure, size_check_procedure,
                               max_error_label, min_error_label):
        type_conversion_procedure(value)
        if size_check_procedure(value) > constraint.max:
            self.errors.append({'field': constraint.name, 'error': max_error_label})
        if size_check_procedure(value) < constraint.min:
            self.errors.append({'field': constraint.name, 'error': min_error_label})


class UnknownFieldType(Exception):
    pass


def payload(payload_template):
    def payload_retriever(func):
        def wrapped(instance, *args, **kwargs):
            try:
                body = cherrypy.request.body.read().decode(DEFAULT_ENCODING)
                payload_json = json.loads(body)
            except ValueError as e:
                return ResponseEntity(status='error', error='Cannot parse request body')

            payload_binder = PayloadBinder(payload_json, payload_template)
            if payload_binder.has_errors():
                return ResponseEntity(status='error', error=payload_binder.errors)
            kwargs.update({'payload': payload_binder.entity})
            return func(instance, *args, **kwargs)
        return wrapped
    return payload_retriever


def content_type(mime):
    def value_retriever(func):
        def wrapped(instance, *args, **kwargs):
            cherrypy.response.headers['Content-Type'] = mime
            return func(instance, *args, **kwargs)
        return wrapped
    return value_retriever


def exception_handler():
    cherrypy.response.status = 500
    cherrypy.response.headers['Content-Type'] = 'application/json'
    cherrypy.response.body = ResponseEntity(status='error', error='Internal error').tojson()


class BaseController(core.BaseBean):
    _cp_config = {'request.error_response': exception_handler,
                  'show_tracebacks': False,
                  'show_mismatched_params': False,
                  'throw_errors': False}

    @cherrypy.expose
    def default(self, *vpath, **params):
        self._setup_default_response_header()
        try:
            method = self._find_handling_method()
            result = self._dispatch_event(method, vpath, params)

            if cherrypy.response.headers['Content-Type'] == 'application/json':
                if result.status == 'error' and cherrypy.response.status != 404:
                    self._set_bad_request_error()
                return result.tojson()
            else:
                return result
        except ValueError as e:
            self._set_bad_request_error()
            core.log.warning(e)
            return ResponseEntity(status='error', error='Bad Request').tojson()

    def _set_bad_request_error(self):
        self._setup_default_response_header()
        cherrypy.response.status = 400

    def _setup_default_response_header(self):
        cherrypy.response.headers['Content-Type'] = 'application/json'

    def _find_handling_method(self):
        return getattr(self, cherrypy.request.method.upper(), None)

    def _dispatch_event(self, method, vpath, params):
        if not method:
            cherrypy.response.status = 404
            return ResponseEntity(status='error', error="Not found")

        return self._call_target_method_with_matching_list_of_arguments(method, params, vpath)

    def _call_target_method_with_matching_list_of_arguments(self, method, params, vpath):
        method_inspector = _method_inspection.MethodInspector(method)
        if method_inspector.contains_args():
            method_inspector.check_if_arguments_match_target_method(params, vpath)
            result = method(*vpath, **params)
        else:
            result = method()
        return result



import inspect


class MethodInspector(object):
    def __init__(self, method):
        self._args = inspect.getargspec(method)

    def contains_args(self):
        return len(self._args.args) > 1 or self._contains_varargs()

    def check_if_arguments_match_target_method(self, args, kwargs):
        if not self._contains_varargs() and (len(args) + len(kwargs) != len(self._arguments_without_special_ones())):
            raise ValueError('Provided arguments does not match target endpoint')

    def _arguments_without_special_ones(self):
        return list(filter(lambda arg: arg not in ['self', 'payload'], self._args.args))

    def _contains_varargs(self):
        return self._args.varargs or self._args.keywords



import redis
import framepy
from framepy import modules
from framepy import _utils

DEFAULT_REDIS_PORT = 6379


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        redis_host = _utils.resolve_property_or_report_error(
            properties=properties,
            key='redis_host',
            log_message='[Redis] Missing redis_host!'
        )
        redis_port = _utils.resolve_property_or_report_error(
            properties=properties,
            key='redis_port',
            log_message='[Redis] Missing redis_port! Setting default value {}'.format(DEFAULT_REDIS_PORT),
            default_value=DEFAULT_REDIS_PORT
        )
        password = _utils.resolve_property_or_report_error(
            properties=properties,
            key='redis_password',
            log_message='[Redis] Missing password! Skipping authentication'
        )

        beans['_redis_pool'] = redis.ConnectionPool(host=redis_host, port=int(redis_port), db=0, password=password)

    def after_setup(self, properties, arguments, context, beans_initializer):
        pass


class ConnectionError(Exception):
    pass


def get_connection(context):
    connection = redis.Redis(connection_pool=context._redis_pool)
    try:
        connection.ping()
    except redis.exceptions.ConnectionError:
        framepy.log.error('[Redis] Connection pool returned invalid connection!')
        raise ConnectionError('Cannot connect to Redis')
    return connection



from framepy import db


def mock_logic(logic_class, context):
    class FakeTransactionContextManager(object):
        def __init__(self):
            pass

        def __enter__(self):
            return None

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    target = logic_class()
    setattr(target, 'context', context)
    db.transaction = lambda _: FakeTransactionContextManager()
    return target



import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.ext.declarative
import contextlib
from framepy import modules
from framepy import _utils

Table = sqlalchemy.ext.declarative.declarative_base()


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        database_url = _utils.resolve_property_or_report_error(
            properties=properties,
            key='database_url',
            log_message='[DB] No database_url found in properties. Skipping ORM engine creation'
        )

        # typically mysql+pymysql://{username}:{password}@url
        beans['db_engine'] = sqlalchemy.create_engine(database_url)
        beans['_session_maker'] = sqlalchemy.orm.sessionmaker(bind=beans['db_engine'])

    def after_setup(self, properties, arguments, context, beans_initializer):
        pass


@contextlib.contextmanager
def transaction(context):
    session = None
    try:
        session = context._session_maker()
        yield session
        session.commit()
    except Exception as e:
        if session is not None:
            session.rollback()
        raise e
    finally:
        if session is not None:
            session.close()


class GenericRepository(object):
    def all(self, transaction):
        return transaction.query(self.__basetable__).all()

    def one(self, transaction, id):
        return transaction.query(self.__basetable__).filter(self.__basetable__.id == id).one()

    def add(self, transaction, entity):
        transaction.add(entity)

    def delete(self, transaction, entity):
        transaction.delete(entity)


class DataAccessException(Exception):
    def __init__(self, message):
        super(DataAccessException, self).__init__(message)
        self.message = message



from framepy import web
from framepy import db
from framepy import core
from framepy import testing
from framepy import beans
from framepy import client
from framepy.core import log, scan_packages, BaseBean
from framepy.web import controller
from framepy.beans import bean

__all__ = ['beans', 'core', 'web', 'db', 'testing', 'client', 'log', 'controller', 'bean', 'scan_packages', 'BaseBean']



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.readlines()

version = '2016.12.148'

setup(
    name='framepy',
    packages=find_packages(exclude=('tests',)),
    version=version,
    description='Very simple web application framework with support for AMQP and DI',
    author='Michal Korman',
    author_email='m.korman94@gmail.com',
    url='https://github.com/mkorman9/framepy',
    download_url='https://github.com/mkorman9/framepy/tarball/{}'.format(version),
    keywords=['web', 'framework', 'amqp', 'di', 'db'],
    classifiers=[],
    install_requires=[requirement for requirement in requirements if len(requirement) > 0]
)



import framepy
import pika
import pika.exceptions
import threading
import time
from framepy import core
from framepy import modules
from framepy import _thread_level_cache
from framepy import _utils

WAIT_TIME_AFTER_CONNECTION_FAILURE = 2
CONNECTION_RETRIES_COUNT = 3
DEFAULT_AMQP_PORT = 5672
CHANNEL_FIELD = 'channel'

annotated_listeners = {}


def listener(queue_name):
    def wrapped(potential_listener_class):
        annotated_listeners[queue_name] = potential_listener_class
        return potential_listener_class
    return wrapped


class Module(modules.Module):
    def before_setup(self, properties, arguments, beans):
        self._map_listeners_from_arguments(arguments)

        broker_address = _utils.resolve_property_or_report_error(
            properties=properties,
            key='broker_address',
            log_message='[AMQP] broker_address not found in properties!',
            default_value=''
        )
        broker_username = _utils.resolve_property_or_report_error(
            properties=properties,
            key='broker_username',
            log_message='[AMQP] broker_username not found in properties!'
        )
        broker_password = _utils.resolve_property_or_report_error(
            properties=properties,
            key='broker_password',
            log_message='[AMQP] broker_password not found in properties!'
        )

        broker_host, broker_port = self._parse_address(broker_address)

        credentials = pika.PlainCredentials(broker_username, broker_password)
        beans['amqp_engine'] = pika.ConnectionParameters(broker_host, broker_port, '/', credentials)

    def after_setup(self, properties, arguments, context, bean_initializer):
        listeners_mappings = arguments.get('listeners_mappings', [])
        for key, bean in annotated_listeners.items():
            listeners_mappings.append(core.Mapping(bean(), key))

        for m in listeners_mappings:
            bean_initializer.initialize_bean('__listener_' + m.path, m.bean, context)
            _register_listener(context, m.path, m.bean.on_message)

    def _map_listeners_from_arguments(self, arguments):
        listeners = arguments.get('listeners') or []
        for listener_class, path in listeners:
            annotated_listeners[path] = listener_class

    def _parse_address(self, broker_address):
        address_parts = broker_address.split(':')

        if len(address_parts) == 1:
            broker_host = broker_address
            broker_port = DEFAULT_AMQP_PORT
        elif len(address_parts) == 2:
            broker_host = address_parts[0]
            broker_port = int(address_parts[1])
        else:
            framepy.log.error('Invalid broker address!')
            return None, None
        return broker_host, broker_port


class BaseListener(core.BaseBean):
    def on_message(self, channel, method, properties, body):
        pass


class ConnectionError(Exception):
    pass


def send_message(context, routing_key, message, durable=True, exchange=''):
    sending_channel = get_channel(context)
    sending_channel.queue_declare(queue=routing_key, durable=durable)
    sending_channel.basic_publish(exchange=exchange,
                                  routing_key=routing_key,
                                  body=message,
                                  properties=pika.BasicProperties(
                                      delivery_mode=2,
                                  ))


def get_channel(context):
    return _thread_level_cache.fetch_from_cache_or_create_new(CHANNEL_FIELD,
                                                              lambda: _establish_connection(context).channel())


def _establish_connection(context):
    new_connection = None
    connection_established = False
    tries_count = CONNECTION_RETRIES_COUNT
    while not connection_established:
        if tries_count:
            try:
                new_connection = pika.BlockingConnection(context.amqp_engine)
                connection_established = True
            except pika.exceptions.ConnectionClosed:
                time.sleep(WAIT_TIME_AFTER_CONNECTION_FAILURE)
                tries_count -= 1
        else:
            framepy.log.error('[AMQP] Cannot establish connection with {0}:{1}'.format(context.amqp_engine.host,
                                                                                       context.amqp_engine.port))
            raise ConnectionError('Cannot establish AMQP connection')
    return new_connection


def _register_listener(context, routing_key, callback):
    thread_local_channel = get_channel(context)

    def receive_action(channel, method, properties, body):
        try:
            callback(channel, method, properties, body)
            channel.basic_ack(delivery_tag=method.delivery_tag)
        except Exception as e:
            framepy.log.error('[AMQP] Error receiving message from queue {0}, exception: {1}'.format(routing_key, e))

    def listener_thread():
        thread_local_channel.queue_declare(queue=routing_key, durable=True)
        thread_local_channel.basic_qos(prefetch_count=1)
        thread_local_channel.basic_consume(receive_action, routing_key)
        thread_local_channel.start_consuming()

    thread = threading.Thread(target=listener_thread)
    thread.daemon = True
    thread.start()
